.list-row.category-list-row {
    border-style: groove;
    border-width: 1pt;
    padding: 10px;
    background-color: @theme_bg_color;
    color: @theme_fg_color;
}

.list-row.category-list-row:selected,
.list-row.category-list-row:selected:focus {
    background-color: @theme_selected_bg_color;
    color: @theme_selected_fg_color;
}

.list-row.category-list-row .category-label {
    padding: 10pt;
}

.banner {
    color: @insensitive_fg_color;
}

.banner-label {
    padding: 10pt;
    font-size: large;
}

.character-list {
    background-color: rgb(255, 255, 255);
}

.character-label {
    font-size: 7em;
    font-weight: bold;
}

.detail-label {
    font-size: small;
}

.related .character {
    font-size: 3em;
}

.related .list-row {
    border-style: groove;
    border-width: 2pt;
}
